/*
 *   OSPFD routing daemon
 *   Copyright (C) 1998 by John T. Moy
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *   
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <tcl.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include "../src/ospfinc.h"
#include "../src/monitor.h"
#include "../src/system.h"
#include "tcppkt.h"
#include "linux.h"
#include "sim.h"
#include "ospfd_sim.h"
#include <time.h>

rtid_t my_id;
SimSys *simsys;
char buffer[MAX_IP_PKTSIZE];
char *LOOPADDR = "127.0.0.1";
char *BROADADDR = "127.255.255.255";

// External declarations
// extern "C" void    srandom(unsigned);

/* The main simulated OSPF loop. Loops getting messages (packets, timer
 * ticks, configuration messages, etc.) and never returns
 * until the simulated OSPF process is told to exit.
 * Syntax:
 *	ospfd_sim "router_id" "controller_port"
 */

int main(int argc, char *argv[])

{
    int n_fd;
    fd_set fdset;
    fd_set wrset;
    uns16 controller_port;
    char temp[30];
    int stat;
    int ctl_fd;

    // Get command line arguments
    if (argc != 3) {
        printf("syntax: ospfd_mon $router_id $controller_port\n");
	exit(1);
    }
    my_id = ntoh32(inet_addr(argv[1]));
    controller_port = atoi(argv[2]);

    // Connect to simulation controller
    // Keep on trying if we're getting timeouts
    do {
        sockaddr_in ctl_addr;
	if (!(ctl_fd = socket(AF_INET, SOCK_STREAM, 0))) {
	    perror("create control socket");
	    exit(1);
	}

	memset((char *) &ctl_addr, 0, sizeof(ctl_addr));
	ctl_addr.sin_family = AF_INET;
	ctl_addr.sin_addr.s_addr = inet_addr(LOOPADDR);
	ctl_addr.sin_port = hton16(controller_port);
	stat = connect(ctl_fd, (struct sockaddr *) &ctl_addr,sizeof(ctl_addr));
	if (stat < 0)
	    close(ctl_fd);
    } while (stat < 0);

    // Create simulation environment
    sys = simsys = new SimSys(ctl_fd);
    // Start OSPF
    ospf = new OSPF(my_id);
    // Log command arguments
    sprintf(temp, "invoked: ospfd_sim %s %s", argv[1], argv[2]);
    simsys->sys_spflog(ERR_SYS, temp);

    // Main loop, never exits
    while (1) {
        bool ifc_data=false;
	FD_ZERO(&fdset);
	FD_ZERO(&wrset);
	// Flush any logging messages
	if (ospf)
	    ospf->logflush();
	// Add connection to controller
	n_fd = simsys->ctl_fd;
	FD_SET(simsys->ctl_fd, &fdset);
	if (simsys->ctl_pkt.xmt_pending())
	    FD_SET(simsys->ctl_fd, &wrset);
	// Add our private port
	n_fd = MAX(n_fd, simsys->uni_fd);
	FD_SET(simsys->uni_fd, &fdset);
	// Add monitoring connection
	simsys->mon_fd_set(n_fd, &fdset, &wrset);
	// Wait for I/O
	if (select(n_fd+1, &fdset, &wrset, 0, 0) < 0)
	    simsys->halt(1, "select failed");
	// Process received data
	if (FD_ISSET(simsys->ctl_fd, &wrset))
	    simsys->ctl_pkt.sendpkt();
	// Received OSPF packets?
	if (FD_ISSET(simsys->uni_fd, &fdset)) {
	    ifc_data = true;
	    simsys->process_uni_fd();
	}
	// Instructions from controller?
	if (!ifc_data && FD_ISSET(simsys->ctl_fd, &fdset))
	    simsys->recv_ctl_command();
	// Process monitor queries and responses
	simsys->process_mon_io(&fdset, &wrset);
    }
    exit(0);
}

/* Process packets received on our private socket.
 * These are packets that have been sent directly
 * to us. If the destination address belongs to us,
 * forward it to the demux. Otherwise, do a routing lookup
 * on the packet and forward it further.
 */

void SimSys::process_uni_fd()

{
    int plen;
    sockaddr_in addr;
    socklen fromlen=sizeof(addr);
    InPkt *pkt;
    InAddr daddr;
    char temp[80];
    in_addr addtmp;
    int phyint;

    plen = recvfrom(uni_fd, buffer, sizeof(buffer),
		    0, (sockaddr *)&addr, &fromlen);
    if (plen < 0) {
        perror("recvfrom");
	return;
    }

    pkt = (InPkt *) (buffer+sizeof(uns32));
    phyint = ntoh32(*((uns32 *) buffer));
    daddr = ntoh32(pkt->i_dest);
    if (!IN_CLASSD(daddr)) {
	InAddr home;
        if (!get_port_addr(daddr, home)) {
	    addtmp.s_addr = pkt->i_dest;
	    sprintf(temp, "Can't forward %s, unknown", inet_ntoa(addtmp));
	    sys_spflog(ERR_SYS, temp);
	}
	else if (home != ospf->my_id()) {
	    MPath *mpp;
	    if ((mpp = ospf->ip_lookup(daddr)) == 0) {
	        addtmp.s_addr = pkt->i_dest;
		sprintf(temp, "Can't forward %s, unreachable",
			inet_ntoa(addtmp));
		sys_spflog(ERR_SYS, temp);
	    }
	    else
	        sendpkt(pkt, mpp->NHs[0].phyint, mpp->NHs[0].gw);
	}
	else
	    local_demux(phyint, pkt, plen);
    }
    else
	local_demux(phyint, pkt, plen);
}

/* Have received an IP packet that is addressed to
 * us (may be a multicast).
 * Fir determine if we are a member of the multicast group,
 * and then dispatch on protocol number.
 */

void SimSys::local_demux(int phyint, InPkt *pkt, int plen)

{
    InAddr dest;
    PhyintMap *phyp;

    if (!(phyp = (PhyintMap *)port_map.find(phyint,0)))
        return;

    dest = ntoh32(pkt->i_dest);
    if (IN_CLASSD(dest)) {
        if (!phyp->promiscuous && !membership.find(dest, phyint))
	    return;
    }

    switch(pkt->i_prot) {
      case PROT_OSPF:
	ospf->rxpkt(phyint, pkt, plen);
	break;
      default:
	break;
    }
}


/* Initialize the Simulation environment.
 * Connect to the controller
 * Start the random number generator.
 */

SimSys::SimSys(int fd) : Linux(0), ctl_pkt(fd)

{
//  rlimit rlim;
    sockaddr_in addr;
    socklen size;
    SimHello hello;

    ctl_fd = fd;
    // Initialize time
    ticks = 0;
    // Allow core files
//  rlim.rlim_max = RLIM_INFINITY;
//  (void) setrlimit(RLIMIT_CORE, &rlim);

    // Open our unicast socket, where people send packets
    // addressed to us (unicasts and multicasts)
    uni_fd = socket(AF_INET, SOCK_DGRAM, 0);
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(uni_fd, (struct sockaddr *) &addr, sizeof(addr)) < 0)
	perror("bind fails");
    size = sizeof(addr);
    if (getsockname(uni_fd, (struct sockaddr *) &addr, &size) < 0) {
	perror("getsockname");
	exit (1);
    }
    uni_port = ntohs(addr.sin_port);

    // Identify ourselves to the server
    hello.rtrid = my_id;
    hello.myport = hton16(uni_port);
    ctl_pkt.queue_xpkt(&hello, SIM_HELLO, 0, sizeof(hello));

    // Open monitoring listen socket
    monitor_listen();
    // Start random number generator
    srandom(getpid());
}

/* Destructor not expected to be called during the life
 * of the program.
 */

SimSys::~SimSys()

{
}

/* Process command received over the controller
 * connection.
 */

void SimSys::recv_ctl_command()

{
    uns16 type;
    uns16 subtype;
    int nbytes;
    char *msg;
    char *end;

    nbytes = ctl_pkt.receive((void **)&msg, type, subtype);
    if (nbytes < 0)
	simsys->halt(1, "Lost controller connection");
    else if (type != 0) {
        AddrMap *addrmap;
        end = msg + nbytes;
        switch (type) {
	  case SIM_TICK:
	    // Advance time
	    ticks++;
	    etime.sec = ticks/TICKS_PER_SECOND;
	    etime.msec = (ticks%TICKS_PER_SECOND) * 1000/TICKS_PER_SECOND;
	    // Process any pending timers
	    ospf->tick();
	    // Send output packets
	    sendqueue();
	    // Send tick reponse
	    send_tick_response();
	    break;
          case SIM_CONFIG:
          case SIM_CONFIG_DEL:
	    config(type, subtype, msg);
	    break;
	  case SIM_ADDRMAP:
	    addrmap = (AddrMap *) msg;
	    for (; (char *)(addrmap + 1) <= end; addrmap++) {
	        AddressMap *entry;
		InAddr addr;
		InAddr home;
		in_addr temp;
		addr = addrmap->addr;
		home = addrmap->home;
	        if (!(entry = (AddressMap *) address_map.find(addr, home))) {
		    entry = new AddressMap(addr, home);
		    address_map.add(entry);
		}
		entry->port = (uns16) addrmap->port;
		entry->home = addrmap->home;
		temp.s_addr = hton32(addr);
	    }
	  default:
	    break;
      }
    }
}

/* Configuration request. Dispatch to the correct routine
 * to do specific processing.
 */

void SimSys::config(int type, int subtype, void *msg)

{
    int status;
    
    status = (type == SIM_CONFIG) ? ADD_ITEM : DELETE_ITEM;

    switch(subtype) {
        PhyintMap *phyp;
	CfgIfc *ifcmsg;
      case CfgType_Gen:
	ospf->cfgOspf((CfgGen *) msg);
	break;
      case CfgType_Area:
	ospf->cfgArea((CfgArea *)msg, status);
	break;
      case CfgType_Range:
	ospf->cfgRnge((CfgRnge *)msg, status);
	break;
      case CfgType_Host:
	ospf->cfgHost((CfgHost *)msg, status);
	break;
      case CfgType_Ifc:
	ifcmsg = (CfgIfc *)msg;
	if (!(phyp = (PhyintMap *)port_map.find(ifcmsg->phyint, 0))) {
	    phyp = new PhyintMap(ifcmsg->phyint);
	    port_map.add(phyp);
	}
	ospf->cfgIfc(ifcmsg, status);
	break;
      case CfgType_VL:
	ospf->cfgVL((CfgVL *)msg, status);
	break;
      case CfgType_Nbr:
	ospf->cfgNbr((CfgNbr *)msg, status);
	break;
      case CfgType_Route:
	ospf->cfgExRt((CfgExRt *)msg, status);
	break;
      case CfgType_Key:
	ospf->cfgAuKey((CfgAuKey *)msg, status);
	break;
      default:
	break;
    }
}

/* Constructor for the physical interface.
 */

PhyintMap::PhyintMap(int phyint) : AVLitem(phyint,0)

{
    working = true;
    promiscuous = false;
    sprintf(name, "N%d", phyint);
}

/* Send a tick response, listing the current state
 * of the link-state database.
 */

void SimSys::send_tick_response()

{
    DBStats *statp;
    AreaIterator iter(ospf);
    SpfArea *ap;
    SpfArea *low;
    int mlen;

    low = 0;
    statp = new DBStats;
    mlen = sizeof(DBStats);
    statp->n_exlsas = ospf->n_extLSAs();
    statp->ex_dbxsum = ospf->xsum_extLSAs();


    while ((ap = iter.get_next())) {
      if (ap->n_active_if == 0)
	  continue;
      if (!low || low->id() > ap->id())
	  low = ap;
    }

    if (low) {
	statp->area_id = low->id();
	statp->n_lsas = low->n_LSAs();
	statp->dbxsum = low->database_xsum();
    }
    else {
	statp->area_id = 0;
	statp->n_lsas = 0;
	statp->dbxsum = 0;
    }

    ctl_pkt.queue_xpkt(statp, SIM_TICK_RESPONSE, 0, mlen);
    delete statp;
}

/* After receiving a tick, send the entire queue
 * of packets. For each packet, must determine the
 * socket and address to send on.
 */

void SimSys::sendqueue()

{
    SimPktQ *qptr;
    SimPktQ *next;

    for (qptr = xmt_head; qptr; qptr = next) {
	InPkt *pkt;
	int phyint;
	InAddr gw;
	InAddr nh;
	size_t len;
	byte *data;

        next = qptr->next;

	data = qptr->ip_data;
	pkt = (InPkt *) (data+sizeof(uns32));
	phyint = qptr->hdr.phyint;
	gw = qptr->hdr.nh;
	delete qptr;

	// Calculate next hop	
	if ((nh = hton32(gw)) == 0)
	    nh = pkt->i_dest;
	// Figure out socket destination address
	len = ntoh16(pkt->i_len);
	*((uns32 *) data) = hton32(phyint);
	if (IN_CLASSD(ntoh32(nh)))
	    send_multicast(data, len);
	else {
	    uns16 port;
	    InAddr owner;
	    sockaddr_in to;
	    if (!(port = get_port_addr(ntoh32(nh), owner))) {
	        char temp[80];
		in_addr addtmp;
		addtmp.s_addr = nh;
		sprintf(temp, "Can't resolve nh %s", inet_ntoa(addtmp));
		sys_spflog(ERR_SYS, temp);
		delete [] ((byte *) pkt);
		continue;
	    }
	    to.sin_family = AF_INET;
	    to.sin_addr.s_addr = inet_addr(LOOPADDR);
	    to.sin_port = hton16(port);

	    if (sendto(uni_fd, data, len+sizeof(uns32), 0,
		       (sockaddr *) &to, sizeof(to)) == -1)
	      perror("sendto");
	    //sys_spflog(ERR_SYS, "sendto failed");
	}

	delete [] data;
    }
    xmt_head = 0;
    xmt_tail = 0;
}

/* Send a multicast packet. Find all the routers attached
 * to the segment and broadcast to them individually.
 */

void SimSys::send_multicast(byte *data, size_t len)

{
    int phyint;
    AddressMap *mapp;
    AVLsearch iter(&address_map);

    phyint = ntoh32(*((uns32 *)data));

    iter.seek(phyint, 0);
    while ((mapp = (AddressMap *) iter.next())) {
	sockaddr_in to;
        if (mapp->index1() != (uns32) phyint)
	    break;
	to.sin_family = AF_INET;
	to.sin_addr.s_addr = inet_addr(LOOPADDR);
	to.sin_port = hton16(mapp->port);

	if (sendto(uni_fd, data, len+sizeof(uns32), 0,
		   (sockaddr *) &to, sizeof(to)) == -1)
	  perror("sendto");
	//sys_spflog(ERR_SYS, "sendto failed");
    }
}

/* Return the port that is listening for unicast
 * packets to te simulated router.
 */

uns16 SimSys::get_port_addr(InAddr addr, InAddr &owner)

{
    AddressMap *mapp;
    AVLsearch iter(&address_map);

    iter.seek(addr, 0);
    if (!(mapp = (AddressMap *) iter.next()) ||
	(mapp->index1() != addr))
        return(0);
    owner = mapp->home;
    return (mapp->port);
}

/* Contructor for an entry in the address map, that maps
 * unicast addresses to group addresses to simulate
 * data-link address translation.
 */

AddressMap::AddressMap(InAddr addr, InAddr home) : AVLitem(addr, home)

{
}
