/*
 *   OSPFD routing daemon
 *   Copyright (C) 1998 by John T. Moy
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation; either version 2
 *   of the License, or (at your option) any later version.
 *   
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *   
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "ospfinc.h"
#include "monitor.h"
#include "system.h"
#include "tcppkt.h"
#include "linux.h"
#include "ospfd_sim.h"
#include "sim.h"


/* Send an OSPF packet out a specific interface.
 * Simply queue the packet to be sent when the
 * next timer tick is received.
 */

void SimSys::sendpkt(InPkt *pkt, int phyint, InAddr gw)

{
    SimPktQ *qptr;
    int len;

    len = ntoh16(pkt->i_len);
    qptr = new SimPktQ;
    qptr->next = 0;
    qptr->hdr.phyint = phyint;
    qptr->hdr.nh = gw;
    qptr->ip_data = new byte[len+sizeof(uns32)];
    // Leave room for phyint number
    memcpy(qptr->ip_data+sizeof(uns32), pkt, len);

    if (xmt_head == 0)
        xmt_head = xmt_tail = qptr;
    else {
        xmt_tail->next = qptr;
	xmt_tail = qptr;
    }
}

/* Send an OSPF packet, interface not specified.
 * This is used for virtual links.
 * Must first do a routing table lookup of the destination
 * to figure out the next hop.
 */

void SimSys::sendpkt(InPkt *pkt)

{
    MPath *mpp;

    // Resolve next hop
    if ((!ipforwarding) ||
	(mpp = ospf->ip_lookup(ntoh32(pkt->i_dest))) == 0) {
	sys_spflog(ERR_SYS, "next hop lookup failed");
	return;
    }
    sendpkt(pkt, mpp->NHs[0].phyint, mpp->NHs[0].gw);
}

/* Return whether or not a physical interface is
 * operational. The phyint of 0, which is illegal,
 * is declared operational so that ASEs can be imported
 * with forwarding addresses of 0.0.0.0.
 */

bool SimSys::phy_operational(int phyint)

{
    PhyintMap *phyp;

    if (phyint == 0)
        return(true);
    phyp = (PhyintMap *) port_map.find(phyint, 0);
    return(phyp && phyp->working);
}

/* Open a network interface for the sending and receiving of OSPF
 * packets. 
 * In the simulator this is a no-op, since all packets are received
 * on a single UDP socket (created during initialization
 */

void SimSys::phy_open(int phyint)

{
}

/* Closing a particular network interface is again a
 * no-op.
 */

void SimSys::phy_close(int phyint)

{
}

/* Join a particular multicast group on a particular interface.
 * Executed when we become DR/Backup, or when an interface first
 * comes up.
 */

void SimSys::join(InAddr group, int phyint)

{
    AVLitem *entry;
    entry = new AVLitem(group, phyint);
    membership.add(entry);
}

/* Leave a particular multicast group, again on a given
 * interface.
 */

void SimSys::leave(InAddr group, int phyint)

{
    AVLitem *entry;
    if ((entry = membership.find(group, phyint)))
        membership.remove(entry);
}

/* Enable or disable IP forwarding.
 */

void SimSys::ip_forward(bool enabled)

{
    ipforwarding = enabled;
}
    
/* Enable multicast forwarding on an interface. This mainly
 * involves putting the interface into promiscuous mode.
 */

void SimSys::enable_multicast_routing(int phyint)

{
    PhyintMap *phyp;
    if ((phyp = (PhyintMap *)port_map.find(phyint,0)))
	phyp->promiscuous = true;
}

/* Disable multicast forwarding on an interface.
 */

void SimSys::disable_multicast_routing(int phyint)

{
    PhyintMap *phyp;
    if ((phyp = (PhyintMap *)port_map.find(phyint,0)))
	phyp->promiscuous = false;
}

/* Add/modify a kernel routing table entry.
 * NULL for simulator, as we just use the routing table
 * entries internal to the OSPF application.
 */

void SimSys::rtadd(InAddr, InMask, MPath *, MPath *, bool)

{
}

/* Delete a kernel routing table entry.
 * Again, NULL for simulator.
 */

void SimSys::rtdel(InAddr, InMask, MPath *)

{
}

/* Add a multicast routing table entry to the kernel.
 */

void SimSys::add_mcache(InAddr, InAddr, int, int *, int, int *)

{
}

/* Delete a multicast routing table entry from the kernel.
 */

void SimSys::del_mcache(InAddr, InAddr)

{
}

/* Return the printable name of a physical interface.
 */

char *SimSys::phyname(int phyint)

{
    PhyintMap *phyp;

    if (!(phyp = (PhyintMap *) port_map.find(phyint, 0)))
        return(0);
    return(phyp->name);
}

/* Print a logging message. Send it over the
 * control connection to the simulation controller.
 */

void SimSys::sys_spflog(int msgno, char *msgbuff)

{
    ctl_pkt.queue_xpkt(msgbuff, SIM_LOGMSG, msgno, strlen(msgbuff)+1);
}

/* Exit the ospfd program, printing a diagnostic message in
 * the process.
 */

void SimSys::halt(int code, char *string)

{   char buffer[80];

    sprintf(buffer, "Exiting: %s", string);
    sys_spflog(ERR_SYS, buffer);
    exit(code);
}
